import os as os
import numpy as np
import random as ran
import typing as typ
import tensorflow as tf
import keras.utils as ut
import keras.layers as lay
import keras.models as mod
import keras.losses as los
import keras.datasets as dt
import keras.optimizers as opt
import matplotlib.pyplot as plt
import keras.activations as act
import sklearn.model_selection as ms

class AEDN:
    def __init__(self,
                 nConvolution:list[int],
                 RandomState:typ.Union[int, None]=None):
        self.nConvolution = nConvolution
        self.RandomState = RandomState
        self.nConvolutionT = nConvolution[::-1]
        plt.style.use('ggplot')
        if self.RandomState is not None:
            ran.seed(self.RandomState)
            np.random.seed(self.RandomState)
            tf.random.set_seed(self.RandomState)
            os.environ['PYTHONHASHSEED'] = str(self.RandomState)
    def Create(self,
               trX:np.ndarray):
        _, self.h, self.w, self.c = trX.shape
        self.InputShape = (self.h, self.w, self.c)
        self.Model = mod.Sequential(name='Autoencoder')
        self.Model.add(lay.InputLayer(input_shape=self.InputShape))
        for i in self.nConvolution:
            self.Model.add(lay.Conv2D(filters=i,
                                      kernel_size=(3, 3),
                                      padding='same'))
            self.Model.add(lay.LeakyReLU(alpha=0.2))
            self.Model.add(lay.MaxPooling2D(pool_size=(2, 2),
                                            padding='same'))
        for i in self.nConvolutionT:
            self.Model.add(lay.Conv2DTranspose(filters=i,
                                               kernel_size=(3, 3),
                                               strides=(2, 2),
                                               padding='same'))
            self.Model.add(lay.LeakyReLU(alpha=0.2))
        self.Model.add(lay.Conv2D(filters=self.c,
                                  kernel_size=(3, 3),
                                  padding='same'))
        self.Model.add(lay.Activation(activation=act.sigmoid))
    def Compile(self,
                Optimizer:opt.Optimizer,
                Loss:los.Loss):
        self.Optimizer = Optimizer
        self.Loss = Loss
        self.Model.compile(optimizer=self.Optimizer,
                           loss=self.Loss)
    def Summary(self):
        print('_' * 60)
        print('Model Summary:')
        self.Model.summary()
        print('_' * 60)
    def Plot(self):
        ut.plot_model(self.Model,
                      to_file='Model.png',
                      show_shapes=True,
                      dpi=1536)
    def Fit(self,
            trX:np.ndarray,
            trY:np.ndarray,
            vaX:np.ndarray,
            vaY:np.ndarray,
            nEpoch:int,
            sBatch:int):
        self.nEpoch = nEpoch
        self.sBatch = sBatch
        self.History = self.Model.fit(x=trX,
                                      y=trY,
                                      batch_size=sBatch,
                                      epochs=nEpoch,
                                      validation_data=(vaX, vaY),
                                      shuffle=True).history
    def Predict(self,
                X:np.ndarray) -> np.ndarray:
        P = self.Model.predict(X, verbose=0)
        return P
    def LossPlot(self):
        T = np.arange(start=1,
                      stop=self.nEpoch + 1,
                      step=1)
        plt.plot(T,
                 self.History['loss'],
                 ls='-',
                 lw=1.2,
                 c='teal',
                 label='Train')
        plt.plot(T,
                 self.History['val_loss'],
                 ls='-',
                 lw=1.2,
                 c='crimson',
                 label='Validation')
        plt.title('Model Loss Plot Over Training Epochs')
        plt.xlabel('Epoch')
        plt.ylabel('Loss')
        plt.yscale('log')
        plt.legend()
        plt.show()
    def PlotResults(self,
                    X:np.ndarray,
                    Y:np.ndarray):
        P = self.Predict(X)
        nD = X.shape[0]
        I = np.arange(start=0, stop=nD, step=1)
        np.random.shuffle(I)
        for i in I:
            e1 = 100 * np.power(Y[i] - X[i], 2).mean() ** 0.5
            e2 = 100 * np.power(Y[i] - P[i], 2).mean() ** 0.5
            plt.subplot(1, 3, 1)
            plt.imshow(Y[i], cmap='gray')
            plt.title('Main Image')
            plt.xlabel('NRSE: 0 %')
            plt.subplot(1, 3, 2)
            plt.imshow(X[i], cmap='gray')
            plt.xlabel(f'NRSE: {e1:.2f} %')
            plt.title('Noised Image')
            plt.subplot(1, 3, 3)
            plt.imshow(P[i], cmap='gray')
            plt.xlabel(f'NRSE: {e2:.2f} %')
            plt.title('Denoised Image')
            plt.show()

sVa = 0.2
nConvolution = [64, 32]
RandomState = 0
Optimizer = opt.Adam()
Loss = los.BinaryCrossentropy()
nEpoch = 10
sBatch = 128

(trvaY, _), (teY, _) = dt.mnist.load_data()

print(f'trvaY: {trvaY.shape}')
print(f'teY: {teY.shape}')

trvaY = np.expand_dims(trvaY, axis=-1)
teY = np.expand_dims(teY, axis=-1)

print(f'trvaY: {trvaY.shape}')
print(f'teY: {teY.shape}')

print(f'Min: {trvaY.min()}')
print(f'Max: {trvaY.max()}')

trvaY = trvaY / 255
teY = teY / 255

print(f'Min: {trvaY.min()}')
print(f'Max: {trvaY.max()}')

print(f'dtype: {trvaY.dtype}')
print(f'trvaY Size: {trvaY.size * trvaY.itemsize}')

trvaY = trvaY.astype('float32')
teY = teY.astype('float32')

print(f'dtype: {trvaY.dtype}')
print(f'trvaY Size: {trvaY.size * trvaY.itemsize}')

trvaX = trvaY + np.random.normal(loc=0, scale=0.3, size=trvaY.shape)
teX = teY + np.random.normal(loc=0, scale=0.3, size=teY.shape)

trvaX = np.clip(a=trvaX, a_min=0, a_max=1)
teX = np.clip(a=teX, a_min=0, a_max=1)

trX, vaX, trY, vaY = ms.train_test_split(trvaX,
                                         trvaY,
                                         test_size=sVa,
                                         random_state=0,
                                         shuffle=True)

nDtr = trX.shape[0]
I = np.arange(start=0, stop=nDtr, step=1)
np.random.shuffle(I)

for i in I[:2]:
    plt.subplot(1, 2, 1)
    plt.imshow(trY[i], cmap='gray')
    plt.title('Main Image')
    plt.subplot(1, 2, 2)
    plt.imshow(trX[i], cmap='gray')
    plt.title('Noised Image')
    plt.show()


DN = AEDN(nConvolution, RandomState=RandomState)

DN.Create(trX)

DN.Summary()

DN.Plot()

DN.Compile(Optimizer, Loss)

DN.Fit(trX, trY, vaX, vaY, nEpoch, sBatch)

DN.LossPlot()

DN.PlotResults(teX, teY)
